#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include "config.h"
#include "user.h"
#include "misc.h"
#include "send.h"
#include "server.h"

GateListType *GatePtrs;
int NumOfGateSocks = 0;

void GateConnect(GateListType *GatePtr);
void GateCheck_CheckTooMany();

void BeginGateScan(UserListType UserNode, char *IP)
{
   GateListType *GatePtr;

   
  if(!GATECHECK)
     return;
/*  Debug(DBGINFO, "BeginGateScan(%s, %s)", Nick, IP);
*/
  if(!strcmp(IP, "127.0.0.1")) /* users on local servers */
      return;
  if(!strcmp(IP, "0.0.0.0")) /* services */
    return;
  if(UserNode == NULL)
  {
     Debug(DBGNOTICE, "BeginGateScan(): Null user ptr");
     return;
  }
  if(!(GatePtr=(GateListType *) malloc(sizeof(*GatePtr))))
  {
      Debug(DBGSEVERE, "Out of memory in GateCheck!");
      return;
  }
  GatePtr->delete = FALSE;
  GatePtr->sock = -1;
  GatePtr->timestamp = time(NULL);
  GatePtr->state = WAIT_TO_CONNECT;
  strcpy(GatePtr->nick, UserNode->Nick);
  strcpy(GatePtr->iphost, IP);
  GatePtr->next = GatePtrs;
  GatePtrs = GatePtr;

  if(NumOfGateSocks > MAXGATESOCKS)
  {
    GatePtr->state = WAIT_TOO_MANY;
    return;
  }
  else
    GateConnect(GatePtr);
}

void GateConnect(GateListType *GatePtr)
{
   struct sockaddr_in GateAddr;
   unsigned long int longip;
   GateListType *PrevGate, *NextGate;
 
  longip = inet_addr(GatePtr->iphost);
  bzero((char *) &GateAddr, sizeof(GateAddr));
  GateAddr.sin_addr.s_addr      = longip;
  GateAddr.sin_family           = AF_INET;
  GateAddr.sin_port             = htons(SOCKSPORT);

  if((GatePtr->sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
  {
     Debug(DBGWARNING, "ERROR: socket() returned error for host: %s.  :%s:",
                       GatePtr->iphost, strerror(errno));
     Debug(DBGWARNING, "Terminating gate checking due to unrecoverable error");
     GATECHECK = FALSE;
     GatePtr->delete = TRUE;
     return;
  }
  NumOfGateSocks++;  /* Only set this once we have established a socket */
  if(GatePtr->sock > MaxSock)
         MaxSock = GatePtr->sock;
  fcntl(GatePtr->sock, F_SETFL, O_NONBLOCK);
  if(connect(GatePtr->sock, (struct sockaddr *) &GateAddr, sizeof(GateAddr)) >= 0)
  {
     /*Debug(DBGINFO, "Called connect() for gate.."); */
     GatePtr->state = WAIT_TO_SEND;
  }
  else
  {
    if(errno == EINPROGRESS)
    {
       /* connect is non blocking, use select to finish..*/
       GatePtr->state = WAIT_TO_CONNECT;
    /*   Debug(DBGINFO, "Non blocking connect to %s, waiting to finish..", GatePtr->iphost); */
    }
    else
    {
    /*   Debug(DBGNOTICE, "Connect failed to %s.  :%s:", GatePtr->iphost, strerror(errno)); */
       GatePtr->delete = TRUE;
       close(GatePtr->sock);
       NumOfGateSocks--;
    }
  }
/* Delete marked gates.. */
  return;
  for(GatePtr=GatePtrs;NextGate;GatePtr=NextGate)
  {
     NextGate=GatePtr->next;
     if(GatePtr->delete)
     {
      /*  Debug(DBGNOTICE, "Removing %s from gate list.", GatePtr->iphost); */
        if(GatePtr == GatePtrs)
          GatePtrs = GatePtr->next;
        else
          PrevGate->next = GatePtr->next;
        free(GatePtr);
     }
     else
        PrevGate = GatePtr;
  }
  GateCheck_CheckTooMany();
}

void GateCheck_CheckConnect(GateListType *GatePtr)
{
/******8<*****8<***********
     After select indicates writability,
     use getsockopt(2) to read the SO_ERROR  option  at
     level SOL_SOCKET to determine whether connect com
     pleted successfully (SO_ERROR is zero)  or  unsuc
     cessfully  (SO_ERROR  is  one  of  the usual error
     codes listed above, explaining the reason for  the
     failure).
*******8<*****8<***********/
  unsigned int opt_val = 0;
  unsigned int opt_size;
  GateListType *PrevGate, *NextGate;

  opt_size = sizeof(opt_val);
 
  if(GatePtr->state == WAIT_TOO_MANY)
      return;
  if(GatePtr->sock < 0)
  {
      Debug(DBGERROR, "GatePtr->sock is negitive at the top of CheckConnect()");
  }
  if(!getsockopt(GatePtr->sock, SOL_SOCKET, SO_ERROR, &opt_val, &opt_size))
  {
      if(opt_val == 0)
      {
       /*  Debug(DBGINFO, "NONBLOCK connect to %s succeded.", GatePtr->iphost); */
         GatePtr->state = WAIT_TO_SEND;
      }
      else
      {
         /*Debug(DBGINFO, "NONBLOCK connect to %s failed. opt_val is: %s", 
                         GatePtr->iphost, strerror(opt_val)); */
         GatePtr->delete = TRUE;
         close(GatePtr->sock);
         NumOfGateSocks--;
      }
  }
  else
  {
     /* Debug(DBGINFO, "getsockopt() failed.. :%s:", strerror(errno)); */
      GatePtr->delete = TRUE;
      close(GatePtr->sock);
      NumOfGateSocks--;
  }
/* Delete marked gates.. */
  return;
  for(GatePtr=GatePtrs;GatePtr;GatePtr=NextGate)
  {
     NextGate=GatePtr->next;
     if(GatePtr->delete)
     {
       /* Debug(DBGNOTICE, "Removing %s from gate list.", GatePtr->iphost); */
        if(GatePtr == GatePtrs)
          GatePtrs = GatePtr->next;
        else
          PrevGate->next = GatePtr->next;
        free(GatePtr);
     }
     else
        PrevGate = GatePtr;
  }
  GateCheck_CheckTooMany();
}

/* TODO : put the destination socks IP in the conf file, and grab the
 * userid from the running x2 account.. use them instead of
 * 199.2.210.241/rubin.
 */
void GateCheck_CheckSend(fd_set fdset)
{
 /* Got some docs on this key..
  * http://www.socks.nec.com/protocol/socks4.protocol
  *  DESC: VN   CD   DSTPORT   DSTIP             USERID ... NULL
  * BYTES:  1    1     2         4               variable
  * VALUES: 4,   1,  26, 11,  199,2,210,241,        "Rubin", 0
  * these are the base 10 values.. note that 199.2.210.241 is
  * agora's ip address. When agora goes down the connections
  * started to fail, due to no route to host.. and the scan 
  * broke. So it would be better to connect to something
  * more configurable/reliable. 
  */
/*unsigned char keystring[] = {0x04, 0x01, 0x1A, 0x0B, 0xC7,
                               0x02, 0xD2, 0xF1, 0x62, 0x6C, 
 *converting to dec..*         0x61, 0x68, 0x74, 0x65, 0x73, 0x74, 0x00};
*/
/*  NOTE: If the IP below (206.58.8.1) is not routable, the socks check will FAIL! */
/* TODO : Put this IP in the config file somehow. */
/*                                  {PORT}  {Agora's IP}   {Userid..} */
  unsigned char keystring[] = {4, 1, 26,11, 206,58,8,1, 'r','u','b','i','n'};

  GateListType *GatePtr, *NextGate, *PrevGate = NULL;
  
  for(GatePtr = GatePtrs;GatePtr;GatePtr=GatePtr->next)
  {
     if(GatePtr->state == WAIT_TOO_MANY || GatePtr->delete)
       continue;
     if(GatePtr->sock < 0)
     {
        Debug(DBGERROR, "sock is negitive in CheckSend for loop");
     }
     if(GatePtr->state == WAIT_TO_CONNECT)
     {                                    
        if(FD_ISSET(GatePtr->sock, &fdset))
        {
          GateCheck_CheckConnect(GatePtr);
          continue;
        }
        continue;
     }
     if(GatePtr->state != WAIT_TO_SEND)
       continue;
     if(FD_ISSET(GatePtr->sock, &fdset))
     {
      /*  Debug(DBGINFO, "Sending magic string to %s", GatePtr->iphost); */
        send(GatePtr->sock, keystring, 17, 0);
        GatePtr->timestamp = time(NULL);
        GatePtr->state = WAIT_TO_READ;
     }
     else
     {
        if(time(NULL) - GatePtr->timestamp > GATECHECKSENDTIMEOUT)
        {
           /* We waited long enough.. cant send... give up. */
           close(GatePtr->sock);
           NumOfGateSocks--;
           GatePtr->delete = TRUE; /* Cant delete now, the list is being looped though. */
        }
     }
  }
/* Delete marked gates.. */
  for(GatePtr=GatePtrs;GatePtr;GatePtr=NextGate)
  {
     NextGate=GatePtr->next;
     if(GatePtr->delete)
     {
        /*Debug(DBGNOTICE, "Removing %s from gate list.", GatePtr->iphost); */
        if(GatePtr == GatePtrs)
          GatePtrs = GatePtr->next;
        else
          PrevGate->next = GatePtr->next;
        free(GatePtr);
     }
     else
        PrevGate = GatePtr;
  }
  GateCheck_CheckTooMany();
}

void GateCheck_CheckRead(fd_set fdset)
{
  GateListType *GatePtr, *NextGate, *PrevGate = NULL;
  char decstr[MAXLEN], buf[50];
  int c, i;
  char tmpstr[MAXLEN] = "";
  FILE *GATEFILE;

  for(GatePtr = GatePtrs;GatePtr;GatePtr=GatePtr->next)
  {
     if(GatePtr->state == WAIT_TOO_MANY || GatePtr->delete)
        continue;
     if(GatePtr->state == WAIT_TO_READ && FD_ISSET(GatePtr->sock, &fdset))
     {
        /* Read the incomming response */
        c = read(GatePtr->sock, tmpstr, 10);
        if(c >=0)
        {
          strcpy(decstr, "");
          for(i=0;i<c;i++)
          {
             sprintf(buf, "%d ", tmpstr[i]);
             strcat(decstr, buf);
          }
          Debug(DBGNOTICE, "Got [%s] from %s", decstr, GatePtr->iphost);
/*
          if(decstr[0])
             send_wallops("GateCheck response: [%s] for %s", decstr, GatePtr->nick);
*/
          /* Apparently new versions are out that use 05b as a responce,
           * or maybe because agora is not reachable today.. regardless,
           * expanding the search a bit so the 2nd byte can be 0x5a or 0x5b.
           * TODO : add this
           * 0x04 0x5a 0x57 0x0b : another type of responce gotten..
           * Ok, found some socks protocol docs:
           * http://www.socks.nec.com/protocol/socks4.protocol
           * Description: VN   CD   DSTPORT   DSTIP
           *       Bytes:  1    1      2        4
           * CD is  90=granted  - gline them
           *        91=rejected/failed 
           *        92=rejected due to lack of ident, 
           *        93=rejected coz ident userid's conflicted 
           *  -Rubin */
          if(   ( (tmpstr[0] == 0) || (tmpstr[0] == 4) || (tmpstr[0] == 48)) && 
             (tmpstr[1] == 90) )   /* 90 = granted */
          {   /* We got one! */
            char uhost[MAXLEN];
            char reason[MAXLEN];

            sprintf(uhost, "*@%s", GatePtr->iphost);
            sprintf(reason, "<%s> 1hr ban -- %s[%s] Your Site has an insecure wingate! ACCESS to %s IRC NETWORK is DENIED!  see  %s for more information!",
                               IrcNick, GatePtr->nick, GatePtr->iphost,NETWORK,GATEURL);
            if(GATECHECK == 2)
            {
/*              sendtosock(Sock, ":%s GLINE * +%s 3600 :%s", ServiceName, uhost, reason); */
              send_gline(uhost, 3600, "%s", reason);
              ProcessAddGline(uhost, 3600, reason);
            }
            else if(GATECHECK == 1)
            {
              send_wallops("\002Insecure WinGate detected for %s. GLINE DISABLED, warning only.", GatePtr->nick);
            }
            if((GATEFILE = fopen(GATEFILENAME, "a")))
            {
                fprintf(GATEFILE, "%s - %s (%s)\n", time2str(time(NULL)), GatePtr->iphost, GatePtr->nick);
                fclose(GATEFILE);
            }
            close(GatePtr->sock);
            NumOfGateSocks--;
            GatePtr->delete = TRUE;
          }
          else
          {
             /* Connected but not an insecure gate */
             close(GatePtr->sock);
             NumOfGateSocks--;
             GatePtr->delete = TRUE;
          }
        }
        else
        {
       /*   Debug(DBGNOTICE, "Read error to socket for %s. 0 bytes read.", GatePtr->iphost); */
          close(GatePtr->sock);
          NumOfGateSocks--;
          GatePtr->delete = TRUE;
          return;
        }
     } 
     else
     {
        if(GatePtr->sock < 0)
           Debug(DBGERROR, "Point A: GatePtr->sock is negitive in CheckRead()");
        if(time(NULL) - GatePtr->timestamp > GATECHECKTIMEOUT)
        {
           /* We waited long enough.. no reply.. give up. */
           close(GatePtr->sock);
           NumOfGateSocks--;
           GatePtr->delete = TRUE; /* Cant delete now, the list is being looped though. */
        }
     }
  }
  /* Need to check for and remove any deleted items here... */
  for(GatePtr=GatePtrs;GatePtr;GatePtr=NextGate)
  {
     NextGate=GatePtr->next;
     if(GatePtr->delete)
     {
        /*Debug(DBGNOTICE, "Removing %s from gate list.", GatePtr->iphost); */
        if(GatePtr == GatePtrs)
          GatePtrs = GatePtr->next;
        else
          PrevGate->next = GatePtr->next;
        free(GatePtr);
     } else
        PrevGate = GatePtr;
  }
  GateCheck_CheckTooMany();
}

void GateCheck_CheckTooMany()
{
  GateListType *GatePtr;

  for(GatePtr=GatePtrs;GatePtr;GatePtr=GatePtr->next)
  {
     if(NumOfGateSocks > MAXGATESOCKS)
        return;
     if(GatePtr->state == WAIT_TOO_MANY)
     {
       /* Debug(DBGINFO, "Starting delayed gatecheck of %s", GatePtr->iphost); */
        GateConnect(GatePtr);
     }
  }
}
