
/*
 *    X2  IRC Operator and Channel services
 *    by Rubin (rubin@afternet.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "config.h"
#include "sock.h"
#include "misc.h"
#include "chan.h"
#include "server.h"
#include "user.h"
#include "send.h"

UserListType FindUser(char *UserID, char *Address, UserSegment TheSegment)
{

  int Pos;
  unsigned long Key;
  UserListType TempListPtr;

  Key = Hash(Address);
  Pos = Key % USERSEGSIZE;

  TempListPtr = TheSegment->Table[Pos];
  while (TempListPtr)
  {
    if ((!(strcmp(TempListPtr->Address, Address))) && (!(strcmp(TempListPtr->UserID, UserID))))
      return (TempListPtr);
    else
      TempListPtr = TempListPtr->NextA;
  }
  return (NULL);
}


void UpdateCloneCount(char *UserID, char *Address, int NewCount, UserSegment TheSegment)
{

  int Pos;
  unsigned long Key;
  UserListType TempListPtr;

  Key = Hash(Address);
  Pos = Key % USERSEGSIZE;

  TempListPtr = TheSegment->Table[Pos];
  while (TempListPtr)
  {
    if ((!(strcmp(TempListPtr->UserID, UserID))) && (!(strcmp(TempListPtr->Address, Address))))
      TempListPtr->Clones = NewCount;
    TempListPtr = TempListPtr->NextA;
  }
}

int GetNumOfClones(char *UserID, char *Address)
{

  UserListType UserNode;

  if (!(UserNode = FindUser(UserID, Address, UsersA)))
    return (0);

  return (UserNode->Clones + 1);

}

UserListType FindNumnick(char *Numnick, UserSegment TheSegment)
{
  int Pos;
  unsigned long Key;
  UserListType TempListPtr;

  Key = Hash(Numnick);
  Pos = Key % USERSEGSIZE;

  TempListPtr = TheSegment->Table[Pos];
  while (TempListPtr)
  {
    if (!(strcmp(TempListPtr->Numnick, Numnick)))
    {
      return (TempListPtr);
    }
    else
    {
      TempListPtr = TempListPtr->NextU;
    }
  }
  return (NULL);
}

UserListType FindNick(char *Nick, UserSegment TheSegment)
{
  int Pos;
  unsigned long Key;
  UserListType TempListPtr;

  Key = Hash(Nick);
  Pos = Key % USERSEGSIZE;
  
  if (TheSegment->Table[Pos] != NULL)
  {
    TempListPtr = TheSegment->Table[Pos];
    while (TempListPtr)
    {
      if (!strcasecmp(TempListPtr->Nick, Nick))
      {
	return (TempListPtr);
      }
      else
      {
	TempListPtr = TempListPtr->NextN;
      }
    }
  }
  return (NULL);
}

boolean IsModeK(UserListType uptr)
{
  if(uptr != NULL)
  {
    if (uptr->Modes & UMODEk || strcasecmp(uptr->Nick, IrcNick) == 0)
      return 1;
  }
  return 0;

}

boolean IsOper(UserListType uptr)
{
  if(uptr != NULL)
    return ((UMODEo & uptr->Modes) != 0);
  return (FALSE);
}

boolean IsGodOper(UserListType uptr)
{

  if(uptr != NULL)
    return ((UMODEgod & uptr->Modes) != 0);
  return (FALSE);
}

boolean IsOperByAdd(char *UserID, char *Address)
{

  int Pos;
  unsigned long Key;
  UserListType TempListPtr;

  Key = Hash(Address); 
  Pos = Key % USERSEGSIZE;

  if (UsersA->Table[Pos])
  {
    TempListPtr = UsersA->Table[Pos];
    while (TempListPtr)
    {
      if ((!(strcmp(TempListPtr->UserID, UserID))) && (!(strcmp(TempListPtr->Address, Address))))
      {
	if ((UMODEo & TempListPtr->Modes) != 0)
	  return (true);
      }
      TempListPtr = TempListPtr->NextA;
    }
  }
  return (false);
}

UserListType AddUser(char *Numnick, char *Nick, char *UserID, char *Address, char *RealName, unsigned int IP, char *Serv, unsigned char Modes)
{
  static int UserSegWarning = 1;
  UserListType NewUser, UserNode;
  ServerListType ServPtr;
  unsigned long Key;
  int PosA, PosN;
  int PosU;
  char CleanUserID[12];
  int CloneCount;
  char buff[MAXLEN], reason[MAXLEN];

  Debug(DBGINFO, "[Adding user] numnick:%s nick:%s Address:%s", Numnick, Nick, Address);
  if (FindNick(Nick, UsersN))
  {
    send_wallops("WARNING!!! Nick collision on %s from server %s.  New user NOT added!!", Nick, Serv);
    return(NULL);
  }
  if (UserSegWarning && ((UserCount * 2) > USERSEGSIZE))
  {
    send_wallops("WARNING! SOMEONE MAY WANT TO INCREASE USERSEGSIZE %d TO SOMETHING LIKE %d AND RECOMPILE ME!", USERSEGSIZE, (USERSEGSIZE * 2) - 1);
    UserSegWarning = 0;
  }
  strncpy(CleanUserID, UserID, 11);
  CleanUserID[11] = '\0';
  CloneCount = GetNumOfClones(CleanUserID, Address);
  if (!(NewUser = (UserListType) malloc(sizeof(struct UserStruct))))
     return(NULL);
  if (!(NewUser->Address = (char *) malloc(strlen(Address) + 1)))
    return(NULL);
  if (!(NewUser->Serv = (char *) malloc(strlen(Serv) + 1)))
    return(NULL);
  if (!(NewUser->RealName = (char *) malloc(strlen(RealName) + 1)))
    return (NULL);
  if (CloneCount > 0)
    UpdateCloneCount(CleanUserID, Address, CloneCount, UsersA);
  strcpy(NewUser->UserID, CleanUserID);
  strcpy(NewUser->Address, Address);
  strcpy(NewUser->Serv, Serv);
  strcpy(NewUser->RealName, RealName);
  NewUser->IP = IP;

  strcpy(NewUser->Nick, Nick);
  strcpy(NewUser->Numnick, Numnick);
  NewUser->Modes = Modes;
  NewUser->Clones = CloneCount;

  NewUser->AuthHandles = NULL;
  NewUser->Chans = NULL;
  NewUser->LastInvite = time(NULL);

  Key = Hash(Address); 
  PosA = Key % USERSEGSIZE;

  Key = Hash(Nick);
  PosN = Key % USERSEGSIZE;

  Key = Hash(Numnick);
  PosU = Key % USERSEGSIZE;

  NewUser->NextA = UsersA->Table[PosA];
  UsersA->Table[PosA] = NewUser;

  NewUser->NextN = UsersN->Table[PosN];
  UsersN->Table[PosN] = NewUser;

  NewUser->NextU = UsersU->Table[PosU];
  UsersU->Table[PosU] = NewUser;

  UserCount++;
  ServPtr = GetServer(Serv);
  if (ServPtr)
    ServPtr->NumOfUsers++;
 
  if (LastNickAdded[0] != '\0')
  {
    if (!(UserNode = FindNick(LastNickAdded, UsersN)))
    {
      strcpy(LastNickAdded, Nick);
      return(NewUser);
    }

    if ((UserNode->Clones >= MAXCLONES) && (!(IsOperByAdd(UserNode->UserID, UserNode->Address))))
    {
      switch (CLONERES)
      {
      case 1:
        send_to_user(UserNode, "WARNING: You have more connections than allowed! Please disconnect one of your clients. Thank you.");
	send_wallops("WARNING: %s!%s@%s has more connections than allowed!", UserNode->Nick, UserNode->UserID, UserNode->Address);
	break;
      case 2:
        send_kill(UserNode, "Too many connections! Current clone limit is %d !", MAXCLONES);
	DeleteUser(UserNode, true);
	break;
      case 3:
       {
        char uh[MAXLEN];
        sprintf(uh, "%s@%s", UserNode->UserID, UserNode->Address);
        send_gline(uh, 3600, "Exceeded %d clones", MAXCLONES);
	send_wallops("NOTICE: auto-glining %s@%s for exceeding %d maximum clones.",
		   UserNode->UserID, UserNode->Address, MAXCLONES);
	sprintf(buff, "%s@%s", UserNode->UserID, UserNode->Address);
	sprintf(reason, "Exceeded %d clones", MAXCLONES);
	ProcessAddGline(buff, 3600, reason);
       }
      }
    }
  }
  strcpy(LastNickAdded, Nick);
  return(NewUser);
}

/* Called when a squit happens. */
void DeleteAllUsersByServer(char *ServerName)
{
  int i;
  UserListType UserPtr;
  UserListType TmpPtr;


  for (i = 0; i < USERSEGSIZE; i++)
  {
    UserPtr = UsersN->Table[i];
    while (UserPtr)
    {
      TmpPtr = UserPtr->NextN;
      if (!strcasecmp(UserPtr->Serv, ServerName))
      {
	DeleteUser(UserPtr, FALSE);
      }
      UserPtr = TmpPtr;
    }
  }
}

void AddAuthUserPtr(UserListType uptr, HandleListType hptr)
{
   AuthUserListType auptr;

   /* Add a pointer to uptr under hptr's authusers list */

   auptr = malloc(sizeof(struct AuthUserStruct));

   auptr->uptr = uptr;
   auptr->Next = hptr->AuthUsers;
   hptr->AuthUsers = auptr;  

}

void AddAuthHandlePtr(HandleListType hptr, UserListType uptr, ChannelListType cptr)
{
  AuthHandleListType ahptr;

  /* Add a pointer to handle and channel under uptr's AuthHandle list */

  ahptr = malloc(sizeof(struct AuthHandleStruct));

  ahptr->cptr = cptr;
  ahptr->hptr = hptr;
  ahptr->Next = uptr->AuthHandles;
  uptr->AuthHandles = ahptr;
}

/* Delete an Auth User that points to uptr, from hptr's login list */
void DelAuthUserPtr(UserListType uptr, HandleListType hptr)
{
   AuthUserListType auptr, auptr_prev = NULL, auptr_next;

   for(auptr = hptr->AuthUsers; auptr; auptr = auptr_next)
   {
      auptr_next = auptr->Next;
      if(auptr->uptr == uptr)  /* If this auptr is the one we wnat to delete.. */
      {
          if(auptr_prev)
             auptr_prev->Next = auptr->Next;
          else
             hptr->AuthUsers = auptr->Next;
          free(auptr);
          return;
      }
      auptr_prev = auptr;
   }
}

void DelAuthHandlePtr(HandleListType hptr, UserListType uptr)
{
  AuthHandleListType ahptr, ahptr_prev = NULL, ahptr_next;

  for(ahptr = uptr->AuthHandles; ahptr; ahptr = ahptr_next)
  {
     ahptr_next = ahptr->Next;
     if(ahptr->hptr == hptr)
     {
        if(ahptr_prev)
          ahptr_prev->Next = ahptr->Next;
        else
          uptr->AuthHandles = ahptr->Next;
        return;
     }
     ahptr_prev = ahptr;
  }
}

void DeleteUser(UserListType uptr, boolean sendnotice)
{
  int PosN, PosA;
  unsigned long Key;
  UserListType CurrPtrA, CurrPtrN, CurrPtrU, TailA = NULL, TailN = NULL, TailU = NULL;
  ServerListType ServPtr;
  AuthHandleListType ahptr, ahptr_next;
  int PosU;

  Key = Hash(uptr->Nick);
  PosN = Key % USERSEGSIZE;
  
  if (UsersN->Table[PosN])
  {
    CurrPtrN = UsersN->Table[PosN];
    while (CurrPtrN)
    {
      if (!(strcmp(CurrPtrN->Nick, uptr->Nick)))
	break;
      TailN = CurrPtrN;
      CurrPtrN = CurrPtrN->NextN;
    }
  }
  else
    return;
  if (!CurrPtrN)
    return;

  Key = Hash(CurrPtrN->Address);
  PosA = Key % USERSEGSIZE;

  CurrPtrA = UsersA->Table[PosA];
  while (CurrPtrA)
  {
    if (!(strcmp(CurrPtrA->Nick, uptr->Nick)))
      break;
    TailA = CurrPtrA;
    CurrPtrA = CurrPtrA->NextA;
  }

  Key = Hash(CurrPtrN->Numnick);
  PosU = Key % USERSEGSIZE;

  CurrPtrU = UsersU->Table[PosU];
  while (CurrPtrU)
  {
     if(!(strcmp(CurrPtrU->Nick, uptr->Nick)))
       break;
     TailU = CurrPtrU;
     CurrPtrU = CurrPtrU->NextU;
  }

  if (IsOper(uptr))
    OperCount--;
  RemoveNickFromAllChannels(uptr);
  if (TailN == NULL)
    UsersN->Table[PosN] = CurrPtrN->NextN;
  else
    TailN->NextN = CurrPtrN->NextN;

  if (TailA == NULL)
    UsersA->Table[PosA] = CurrPtrA->NextA;
  else
    TailA->NextA = CurrPtrA->NextA;

  if (TailU == NULL)
    UsersU->Table[PosU] = CurrPtrU->NextU;
  else
    TailU->NextU = CurrPtrU->NextU;

  UserCount--;
  if (!strcmp(LastNickAdded, uptr->Nick))
    LastNickAdded[0] = '\0';
  if (CurrPtrN->Clones > 0)
    UpdateCloneCount(CurrPtrN->UserID, CurrPtrN->Address, CurrPtrN->Clones - 1, UsersA);

  /* Go down the list of channels this user is logged into, and clear them out */
  for(ahptr = uptr->AuthHandles; ahptr; ahptr = ahptr_next)
  {
     /* Remove the pointer back to us, frm the handles list of users */
     DelAuthUserPtr(uptr, ahptr->hptr); /* Delete anything pointing to uptr, from ahptr->hptr */
     /* save the ->Next ptr and free this ahptr */
     ahptr_next = ahptr->Next;
     free(ahptr);
  }
  ServPtr = GetServer(CurrPtrN->Serv);
  if (ServPtr)
    ServPtr->NumOfUsers--;
  free(CurrPtrN->Address);
  free(CurrPtrN->Serv);
  free(CurrPtrN);
}

void KillUserList()
{

  int i;
  UserListType UserList, TempPtr;

  for (i = 0; i < USERSEGSIZE; i++)
  {

    if (UsersA->Table[i])
    {
      UserList = UsersA->Table[i];
      while (UserList)
      {
	TempPtr = UserList;
	UserList = UserList->NextA;
	DeleteUser(TempPtr, false);
      }
      UsersA->Table[i] = NULL;
    }
  }
  UserCount = 0;
}

void SetAway(UserListType uptr)
{
  if(uptr == NULL)
    return;
  uptr->Modes = uptr->Modes | UMODEaway;
}

void ClearAway(UserListType uptr)
{
  if(uptr == NULL)
    return;
  uptr->Modes = uptr->Modes & ~UMODEaway;
}

void ProcessUserMode(UserListType uptr, char *Modes)
{
/* i'm not realy satisfied with the way this works */

  char ch;
  char Operator = '+';
  int len, i;
  unsigned int UserMode;

  if (uptr == NULL)
    return;

  UserMode = NOMODES;
  len = strlen(Modes);
  for (i = 0; i < len; i++)
  {
    ch = Modes[i];
    switch (ch)
    {
    case 'o':
      UserMode = UMODEo;
      break;
    case 'd':
      UserMode = UMODEd;
      break;
    case 'w':
      UserMode = UMODEw;
      break;
    case 's':
      UserMode = UMODEs;
      break;
    case 'i':
      UserMode = UMODEi;
      break;
/* New afternet "GOD" mode +h (helper) */
    case 'h':
      UserMode = UMODEgod;
      break;
/* ircu hack2 notices +g */
    case 'g':
      UserMode = UMODEg;
      break;
    case 'k':
      UserMode = UMODEk;
      break;
    case 'x':
      UserMode = UMODEx;
      break;
    case 'r':
      UserMode = UMODEr;
      break;
    case 'R':
      UserMode = UMODER;
      break;
    case '-':
      Operator = ch;
      break;
    case '+':
      Operator = ch;
      break;
    default:
      Debug(DBGWARNING, "ERROR: Unknown usermode %c by %s.", ch, uptr->Nick);
      break;
    }
    if (Operator == '-')
    {
      uptr->Modes = uptr->Modes & ~UserMode;
      if (UserMode == UMODEo)
      {
	OperCount--;
      }
    }
    else if (Operator == '+')
    {
      uptr->Modes = uptr->Modes | UserMode;
      if (UserMode == UMODEo)
      {
	OperCount++;
        /* Check if the oper is away for .IRCOPS */
	send_whois(uptr);
      }
    }
  }
  if ((UMODEo & uptr->Modes) != 0)
    return;
}

/* TODO - these two only used by ProcessNickChange sofar (rest not implemented) */
UserListType RemoveUserNHash(char *Nick, UserSegment Seg)
{
  UserListType CurrPtrN, TailN = NULL;
  int PosN = Hash(Nick) % USERSEGSIZE;

  CurrPtrN = UsersN->Table[PosN];
  while (CurrPtrN)
  {
     if(!(strcmp(CurrPtrN->Nick, Nick)))
       break;
     TailN = CurrPtrN;
     CurrPtrN = CurrPtrN->NextN;
  }
  if (CurrPtrN == NULL)
  	return NULL;
  	
  if (TailN == NULL)
    UsersN->Table[PosN] = CurrPtrN->NextN;
  else
    TailN->NextN = CurrPtrN->NextN;
    
  return CurrPtrN;
}

void AddUserNHash(UserListType User, UserSegment Seg)
{
	int PosN = Hash(User->Nick) % USERSEGSIZE;
		
	User->NextN = Seg->Table[PosN];
	Seg->Table[PosN] = User;
}

void ProcessNickChange(char *OldNick, char *NewNick)
{
	UserListType UserN;
	if (!(UserN = RemoveUserNHash(OldNick, UsersN))) {
		Debug(DBGWARNING, "NULL tried to change nicks!");
		return;
	}
	strcpy(UserN->Nick, NewNick);
	AddUserNHash(UserN, UsersN);
}


/* EOF */
